;; Generated package description from gtags-mode.el  -*- no-byte-compile: t -*-
(define-package "gtags-mode" "1.6.0.20240511.23214" "GNU Global integration with xref, project and imenu." '((emacs "28")) :commit "b1cd0813f9e5cc5ff9c407c6597663867df42637" :keywords '("xref" "project" "imenu" "gtags" "global") :url "https://github.com/Ergus/gtags-mode")
